/*************************************************************************
 *
 * mod_search.c -- A list searching module for IRMP3
 *
 * Copyright (C) by Christopher Wise <cwwise@yahoo.com>
 *
 * This is an list searching module for IRMP3. It reads in lists derived from
 * id3 tags and searches based on ir remote input. It displays the search
 * hits on the LCD display and loads the selected albums or tracks.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>

#include "tools.h"
#include "config.h"
#include "log.h"
#include "mod.h"
#include "mod_search.h"

/*************************************************************************
 * GLOBALS
 */
#define ARTIST 1
#define ALBUM 2
#define TITLE 3

#define FOWARD 1
#define BACK 2

int mod_search_level;
int mod_search_nummatches;
int mod_search_display;
listitem *mod_search_artist = NULL;	//pointers to search lists
listitem *mod_search_album = NULL;
listitem *mod_search_title = NULL;
listitem *mod_search_list;		//pointer to the main list being searched
listitem *mod_search_matches;		//pointer to a temporary list of matches
listitem *mod_search_displayitem;

/*************************************************************************
 * MODULE INFO
 *	This structure is used to pass to mod_register to register
 *	our module with the main program.
 */
mod_t mod_search = {
	mod_search_deinit,	// deinit function
	mod_search_reload,	// called when got SIGHUP
	0,			// watchfd
	NULL,			// poll
	NULL,			// periodical update
	mod_search_message,	// message handler
};


/*************************************************************************
 * RECEIVE MESSAGE
 *	Whenever another module sends out a message via mod_sendmsg(),
 *	this function of all active modules will be called to notify
 *	them about the new message. msgtype shows the type of
 *	the message, msg contains the message.
 */
void mod_search_message (int msgtype, char *msg)
{
	char *c1, *c2;

	log_printf(LOG_DEBUG, "mod_search_message(): got message type %d: '%s'\n", msgtype, msg);
	//process messages from mod_lirc
	// handle input messages
	if (msgtype == MSGTYPE_INPUT) {
		c1 = msg ? strtok(msg, " \t") : NULL;
		c2 = c1 ? strtok(NULL, " \t") : NULL;

		// process search commands
		if (c1 && c2 && !strcasecmp(c1, "search")) {
			if (!strcasecmp(c2, "artist") && mod_search_artist != NULL)
				mod_search_select(c2, ARTIST);
			else if (!strcasecmp(c2, "album") && mod_search_album != NULL)
				mod_search_select(c2, ALBUM);				
			else if (!strcasecmp(c2, "title") && mod_search_title != NULL)
				mod_search_select(c2, TITLE);
			else if (!strcasecmp(c2, "next"))
				mod_search_showlist(FOWARD);
			else if (!strcasecmp(c2, "previous"))
				mod_search_showlist(BACK);
			else if (!strcasecmp(c2, "accept"))
				mod_search_accept(mod_search_displayitem);
			//any other string is a search string
			else if (mod_search_level>0) 
				mod_search_evaluate(c2, mod_search_level, mod_search_list);
		}
	}
}

/*************************************************************************
 * SELECT A SEARCH TYPE
 * this function also initialises a few associated variables
 */

void mod_search_select(char *search_str, int stype) {
	log_printf(LOG_DEBUG, "mod_search: looking for %s\n", search_str);
	mod_search_display = 0;
	mod_search_level = 1;
	mod_search_nummatches = 0;
	mod_sendmsgf(MSGTYPE_INPUT, "searchmessage Choosing %s", search_str);
	if (mod_search_matches != NULL) {
		mod_search_freelist(mod_search_matches);
		mod_search_matches = NULL;
	}
	switch (stype) {
		case ARTIST: mod_search_list = mod_search_artist; break;
		case ALBUM : mod_search_list = mod_search_album;  break;	
		case TITLE : mod_search_list = mod_search_title;  break;
	}
}

/*************************************************************************
 *  STEP THROUGH CURRENT LIST
 */
void mod_search_showlist(int direction) {
	if (mod_search_matches == NULL || mod_search_nummatches < 1) 
		return;
	//if not displaying list do so
	if (mod_search_display == 0) {
		mod_search_displayitem=mod_search_matches;
		mod_search_display = 1;
		mod_sendmsgf(MSGTYPE_INPUT, "searchmessage %s", mod_search_displayitem->display_str);
		log_printf(LOG_DEBUG, "mod_search: nextitem = %p\n", mod_search_displayitem->next_item);
	}
	else { 
		if (direction == FOWARD && mod_search_displayitem->next_item != NULL)
			mod_search_displayitem = mod_search_displayitem->next_item;
		if (direction == BACK && mod_search_displayitem->previous_item != NULL)
			mod_search_displayitem = mod_search_displayitem->previous_item;
		mod_sendmsgf(MSGTYPE_INPUT, "searchmessage %s", mod_search_displayitem->display_str);
	}

}

/*************************************************************************
 *  ACCEPT CURRENT ITEM
 */
void mod_search_accept(listitem *item) {
	//if displaying the list
	if (mod_search_display == 1 && item != NULL) {
		//add the currently displayed item playlist
		mod_sendmsgf(MSGTYPE_INPUT, "playlist loaddir %s*.mp3", item->path);
		mod_search_display = 0;
		mod_search_level = 0;
	}
}
/*************************************************************************
 * MODULE INIT FUNCTION
 *	This is called right after starting IRMP3 and should set up global
 *	variables and register your module via mod_register().
 *	It should return NULL if everything went ok, or an error string
 *	with the error description.
 */
char *mod_search_init (void)
{
	log_printf(LOG_DEBUG, "mod_search_init(): initializing\n");

	// register our module with the main program so that it knows
	// what we want to handle. We must provide a pointer to a
	// mod_t variable which is valid during the whole runtime, so
	// don't specify a local variable here.
	mod_register(&mod_search);

	//load the artist/album/title list files
	mod_search_artist = mod_search_loadfile(ARTIST);
	mod_search_album = mod_search_loadfile(ALBUM);
//	mod_search_title = mod_search_loadfile(TITLE);

	return NULL;
}

/*************************************************************************
 * 
 */
listitem *mod_search_loadfile(int type) {
	char buffer[256];
	FILE *listfile;
	char *name, *path, *display;
	char *filename;
	listitem *list=NULL;

	switch (type) {
		case ARTIST: 
			filename = config_getstr("search_artistfile", "/etc/irmp3.artists");
			break;
		case ALBUM : 
			filename = config_getstr("search_albumfile", "/etc/irmp3.albums");
			break;
		case TITLE : 
			filename = config_getstr("search_titlefile", "/etc/irmp3.titles");
			break;
	}
	log_printf(LOG_DEBUG, "mod_search_init(): opening %s\n", filename);
	listfile = fopen (filename, "rt");
	if ( !listfile ) {
		log_printf(LOG_DEBUG, "mod_search_init(): Unable to open the file %s\n", filename);
		return NULL;
	}
	while(fgets(buffer, 256, listfile)!=NULL) {
//	    	log_printf(LOG_DEBUG, "mod_search_init(): reading from %s\n", filename);
		name =  buffer ? strtok(buffer, "\t") : NULL;
		path = name ? strtok(NULL, "\t") : NULL;

		display = path ? strtok(NULL, "\t") : NULL;
		list = add_to_list(list, name, path, display);
	        log_printf(LOG_DEBUG, "name=%s path=%s, display=%s\n", name, path, display);
	}
	fclose(listfile);
	return list;
}
/*************************************************************************
 * MODULE DEINIT FUNCTION
 *	This is called right before IRMP3 shuts down. You should clean
 *	up all your used data here to prepare for a clean exit.
 */
void mod_search_deinit (void)
{
	// close list files?
	mod_search_freelist(mod_search_artist);
	log_printf(LOG_DEBUG, "mod_search_message(): freed artists\n");
	mod_search_freelist(mod_search_album);
	log_printf(LOG_DEBUG, "mod_search_message(): freed albums\n");
//	mod_search_freelist(mod_search_title);
//	log_printf(LOG_DEBUG, "mod_search_message(): freed titles\n");

	log_printf(LOG_DEBUG, "mod_search_deinit(): deinitialized\n");
}


/*************************************************************************
 * MODULE RELOAD FUNCTION
 *	This is called whenever IRMP3 needs to reload all configs (usually
 *	when receiving a SIGHUP). You should reinit your module or check
 *	for changed configuration here so that changes take effect.
 */
char *mod_search_reload (void)
{
	log_printf(LOG_DEBUG, "mod_search_reload(): reloading\n");
	mod_search_deinit();
	mod_search_init();
	return NULL;
}

/*************************************************************************
 * 
 */

void mod_search_evaluate (char *search_str, int search_level, listitem *source_list) {
	listitem *current_item;
	char message[200];	
	
	if (search_level == 1) {  //create a temporary list of matches
		mod_search_nummatches=0;
		current_item = source_list;
		while (current_item != NULL) {
			if (strchr(search_str, current_item->name[0]) != NULL) {
				mod_search_matches = add_to_list(mod_search_matches, current_item->name, current_item->path, current_item->display_str);
				log_printf(LOG_DEBUG, "mod_search: added = %s\n", current_item->name);
				mod_search_nummatches++;
			}
			current_item = current_item->next_item;
		}
	}
	else 	//remove non-matches from the temporary list
		mod_search_matches = mod_search_cull(search_str, search_level, mod_search_matches); 
	mod_search_level++;
	log_printf(LOG_DEBUG, "mod_search: %i matches\n", mod_search_nummatches);
	snprintf(message, sizeof(message)-1, "%i matches", mod_search_nummatches);
	mod_sendmsgf(MSGTYPE_INPUT, "searchmessage %s", message);
	return;
}

/*************************************************************************
 *  Add an item to a linked list
. Items are added to the start of the list
 *  When creating a new list, it must be called with start_ptr = NULL
 *  to ensure the list is terminated.
 */

listitem *add_to_list(listitem *start_ptr, char *name, char *path, char *display) 
{
	listitem *new_ptr;
   
	new_ptr = malloc(sizeof(listitem));
	if(new_ptr == NULL) {
		log_printf(LOG_DEBUG, "mod_search: Error, can't allocate memory in mod_search\n");
		return(start_ptr);
	}
	new_ptr->name = strdup(name);
	new_ptr->path = strdup(path);
	new_ptr->display_str = strdup(display);	
	if (start_ptr != NULL)
		start_ptr->previous_item = new_ptr;
	new_ptr->previous_item = NULL;
	new_ptr->next_item = start_ptr;
	return new_ptr;
}


/*************************************************************************
 * deletes items from a linked list
 */

listitem *mod_search_cull(char *search_str, int level, listitem *list) {
	listitem *current_item, *temp_item;

	current_item = list;
	while (current_item != NULL) {
		log_printf(LOG_DEBUG, "mod_search: level = %i, search str = %s\n", level, current_item->name);
		if (strchr(search_str, current_item->name[level-1]) == NULL) {  //discard this item
			mod_search_nummatches--;
			if (current_item->previous_item==NULL) {  //the item is the list header
//				log_printf(LOG_DEBUG, "mod_search: removing = %s\n", current_item->name);
				current_item=current_item->next_item;
				mod_search_freeitem(current_item->previous_item);
				current_item->previous_item = NULL;
				list=current_item;
			}
			else {
//				log_printf(LOG_DEBUG, "mod_search: removing2 = %s\n", current_item->name);
				current_item->previous_item->next_item = current_item->next_item;
				if (current_item->next_item != NULL)
					current_item->next_item->previous_item = current_item->previous_item;
				temp_item = current_item;
				current_item=current_item->next_item;
				mod_search_freeitem(temp_item);
			}
		}
		else {  //keep this item - it matches
			if (current_item!=NULL)
		    		current_item = current_item->next_item;
		}
	}
	return list;
}

/*************************************************************************
 * Free a a listitem structure.
 */
void mod_search_freeitem(listitem *item) {
    free(item->name);
    free(item->path);
    free(item->display_str);
    free(item);
}

/*************************************************************************
 * Free a NULL terminated linked list of listitems
 */

void mod_search_freelist(listitem *list) {
	listitem *temp;
	while (list != NULL) {
		temp = list->next_item;
		mod_search_freeitem(list);
		list = temp;
	}
	return;
}
/*************************************************************************
 * EOF
 */
